/*
  This program is made by Daniel Horchner.
    e-mail: Daniel.Horchner@95.Student.WAU.NL or bodo@worldaccess.nl
  I got the CPU and port info from Ralf Brown's interrupt list version 5.5
  in particular from the files OPCODE.LST and 86BUGS.LST. So credits go to:
  Potemkin's Hackers Group - Buy an english grammar book! ;) (OPCODE.LST)
    e-mail: avp@dvm.msk.ru
  Harald Feldmann from Hamarsoft (86BUGS.LST)
    e-mail: harald.feldmann@almac.co.uk
  Ralf Brown (The interrupt list)
    e-mail: ralf@pobox.com

  The source compiles with TC 3.0 and TASM 4.1. If you don't have a 32-bit
  (inline or external) assembler you could change all the 32-bit mnemonics
  into their 16-bit versions and prefix those mnemonics with the operand-size
  prefix, 66h. For example
    pushfd
  can be changed into
    db 66h
    pushf
  The operand-size prefix makes that the processor sees
    pushfd
  The use of constants is little more complicated. For example
    or eax,200000h
  could be translated to
    mov bx,20h
    db 66h
    shl bx,16
    mov bx,0000h
    db 66h
    or ax,bx
  The processor sees
    mov bx,20h
    shl ebx,16
    mov bx,0000h
    or eax,ebx

  This program will only work on a 100% Cyrix 5x86 compatible processor (it
  must have a CCR3 and a CCR4). It will run if you don't have a 100% Cyrix
  compatible processor but it won't enable or disable the CPUID instruction.
  This program checks first if the processor has the CPUID instruction at all.
  By doing so it enables CPUID if the processor supports it. If the first
  argument is 'enable' it leaves the CPUID instruction enabled. If the first
  argument is 'disable' this program disables the CPUID instruction. If there
  was some other argument it restores the old state of CPUID availability
  before exiting this program.
*/
#pragma inline
#define wait6cycles \
  jmp $+2;	    \
  jmp $+2;	    \
  jmp $+2;	    \
  jmp $+2;	    \
  jmp $+2;	    \
  jmp $+2;	    \

unsigned int CPUIDavailable(void)
{
  asm {
    .386
    pushfd
    pop eax
    or eax,0x200000		// Try to set ID (CPUID available) bit
    push eax
    popfd
    pushfd
    pop eax
    test eax,0x200000           // If ID bit set -> CPUID instruction avail.
    jnz CPUIDenabled
/*
    When this point has been reached, CPUID doesn't seem to be available. But
    some processors (Cyrix) have the (ability to set the) ID bit set off by
    default. To make it possible to set this bit one has to set CPUIDEN
    (CPUID Enable, bit 7) in CCR4 (Configuration Control Register 4). But in
    order to write to CCR4 the MAPEN (Register Mapping Enable) bits in CCR3
    must have the value 0001b (=Enable access to all configuration registers).
*/
    mov al,0xc3                 // c3h = index of CCR3
    out 0x22,al                 // next read/write of port 23h from/to CCR3
    wait6cycles
    in al,0x23                  // al = value of CCR3
    and al,0x0f                 // Clear MAPEN bits
    or al,0x10			// Set MAPEN to 1 -> All registers accessible
    mov ah,al                   // Save future CCR3
    mov al,0xc3
    out 0x22,al
    wait6cycles
    mov al,ah
    out 0x23,al                 // MAPEN (bits 4-7) in CCR3 are set to 0001b
    mov al,0xe8			// e8h = index of CCR4
    out 0x22,al                 // next read/write of port 23h from/to CCR4
    wait6cycles
    in al,0x23                  // al = value of CCR4
    or al,0x80                  // Set CPUIDEN (Enable CPUID instr. , bit 7)
    mov ah,al                   // Save future CCR4
    mov al,0xe8
    out 0x22,al
    wait6cycles
    mov al,ah
    out 0x23,al                 // CPUIDEN (bit 7) in CCR4 is set
    pushfd
    pop eax
    or eax,0x200000		// Try again to set ID bit
    push eax
    popfd
    pushfd
    pop eax
    test eax,0x200000           // If ID bit set -> CPUID instruction avail.
    jnz CPUIDenabled
  }
  return (0);                   // CPUID instruction not available
  CPUIDenabled:
  return (1);                   // CPUID instruction available
}

unsigned int CPUIDdisable(void)
{
  asm {
    pushfd
    pop eax
    or eax,0x200000		// Try to set ID (CPUID available) bit
    push eax
    popfd
    pushfd
    pop eax
    test eax,0x200000           // If ID bit set -> CPUID instruction avail.
    jnz CPUIDenabled
    jmp CPUIDdisabled
  }
  CPUIDenabled:
  asm {
    mov al,0xc3                 // c3h = index of CCR3
    out 0x22,al                 // next read/write of port 23h from/to CCR3
    wait6cycles
    in al,0x23                  // al = value of CCR3
    and al,0x0f                 // Clear MAPEN bits
    or al,0x10			// Set MAPEN to 1 -> All registers accessible
    mov ah,al                   // Save future CCR3
    mov al,0xc3
    out 0x22,al
    wait6cycles
    mov al,ah
    out 0x23,al                 // MAPEN (bits 4-7) in CCR3 are set to 0001b
    mov al,0xe8			// e8h = index of CCR4
    out 0x22,al                 // next read/write of port 23h from/to CCR4
    wait6cycles
    in al,0x23                  // al = value of CCR4
    and al,not 0x80             // Reset CPUIDEN (Enable CPUID instr. , bit 7)
    mov ah,al                   //  -> CPUID is disabled
    mov al,0xe8                 // Save future CCR4
    out 0x22,al
    wait6cycles
    mov al,ah
    out 0x23,al                 // CPUIDEN (bit 7) in CCR4 is cleared
/*                                // Set MAPEN to default:
    mov al,0xc3                 // c3h = index of CCR3
    out 0x22,al                 // next read/write of port 23h from/to CCR3
    wait6cycles
    in al,0x23                  // al = value of CCR3
    and al,0x0f                 // Clear MAPEN bits (default)
    mov ah,al                   // Save future CCR3
    mov al,0xc3
    out 0x22,al
    wait6cycles
    mov al,ah
    out 0x23,al                 // MAPEN (bits 4-7) in CCR3 are set to 0000b
*/
    pushfd
    pop eax
    or eax,0x200000		// Try again to set ID bit
    push eax
    popfd
    pushfd
    pop eax
    test eax,0x200000           // If ID bit set -> CPUID instruction avail.
    jz CPUIDdisabled
  }
  return (0);                   // CPUID instruction still available
  CPUIDdisabled:
  return (1);                   // CPUID instruction not available
}

unsigned int CPUIDenabled(void)
{
  asm {
    pushfd
    pop eax
    or eax,0x200000		// Try to set ID (CPUID available) bit
    push eax
    popfd
    pushfd
    pop eax
    test eax,0x200000           // If ID bit set -> CPUID instruction avail.
    jnz CPUIDenabled
  }
  return (0);                   // CPUID instruction not available
  CPUIDenabled:
  return (1);                   // CPUID instruction available
}

unsigned int main(int n_args, char *arg[])
{
  int oldCPUIDstate;

  if (!strcmp("/?",arg[1]))     // Help screen
  {
    printf("cpuid /?      - Displays this help.\n"
           "cpuid enable  - Enables CPUID instruction.\n"
           "cpuid disable - Disables CPUID instruction.\n"
           "cpuid info    - Displays availability of CPUID.\n");
    return (0);                 // No error -> don't return with error code
  }

  if (CPUIDenabled())           // Save current state of CPUID availability
    oldCPUIDstate=1;
  else
    oldCPUIDstate=0;

  if (!CPUIDavailable())
  {
    printf("The CPUID instruction is not available on this processor.\n");
    exit(1);
  }
  else // CPUID is available on this processor -> do what user asks
    if (!strcmp("enable",arg[1])) // CPUIDavailable() enables CPUID
      printf("The CPUID instruction is enabled.\n");
    else
      if (!strcmp("disable",arg[1]))
      {
        if (!CPUIDdisable())
        {
          printf("Could not disable the CPUID instruction.\n");
          exit(1);
        }
        else
          printf("The CPUID instruction is disabled.\n");
      }
      else                      // arg[1] was not 'enable' or 'disable'
      {
        if (!strcmp("info",arg[1]))
          printf("The CPUID instruction is currently %savailable.\n"
                 "The ID-flag in EFLAGS can %sbe set.\n",
                 !oldCPUIDstate ? "not ": "",!oldCPUIDstate ? "not ": "");
        else
          printf("There must be a valid argument given.\n"
                 "Type 'cpuid /?' for help.\n");
        if (oldCPUIDstate==0)   // If CPUID was not available (it is now) ->
          CPUIDdisable();       //  restore old state
      }
  return (0);                   // No error -> don't return with error code
}

